#include "kgraph.h"
using namespace std;
#include "Ecuildean.h"
#include <cmath>
#include <cstring>
#include <malloc.h>
#include <vector>
#include <fstream>
#include <stdexcept>
#include <boost/assert.hpp>
#include <iostream>
#include <cfloat>

#define KGRAPH_MATRIX_ALIGN 64

template <typename T, unsigned A = KGRAPH_MATRIX_ALIGN>
class Matrix {
    unsigned col;
    unsigned row;
    size_t stride;
    char *data;

    void reset (unsigned r, unsigned c) {
        row = r;
        col = c;
        stride = (sizeof(T) * c + A - 1) / A * A;
        /*
        data.resize(row * stride);
        */
        if (data) free(data);
        data = (char *)memalign(A, row * stride); // SSE instruction needs data to be aligned
        if (!data) throw runtime_error("memalign");
    }
public:
    Matrix (): col(0), row(0), stride(0), data(0) {}
    Matrix (unsigned r, unsigned c): data(0) {
        reset(r, c);
    }
    ~Matrix () {
        if (data) free(data);
    }
    unsigned size () const {
        return row;
    }
    unsigned dim () const {
        return col;
    }
    size_t step () const {
        return stride;
    }
    void resize (unsigned r, unsigned c) {
        reset(r, c);
    }
    T const *operator [] (unsigned i) const {
        return reinterpret_cast<T const *>(&data[stride * i]);
    }
    T *operator [] (unsigned i) {
        return reinterpret_cast<T *>(&data[stride * i]);
    }
    void zero () {
        memset(data, 0, row * stride);
    }
    
    void load (const std::string &file_path) {
        std::cout << "Reading Data: " << file_path << std::endl;
        std::ifstream ifs;
        ifs.open(file_path, std::ios::binary);
        assert(ifs.is_open());
        uint32_t N;  // num of points
        ifs.read((char *)&N, sizeof(uint32_t));
        std::cout << "# of points: " << N << std::endl;
        unsigned num_dimensions = 100;
        reset(N, num_dimensions);
        zero();
        ifs.seekg(sizeof(uint32_t), ios::beg);
        for (unsigned i = 0; i < N; ++i) {
            ifs.read(&data[stride * i], sizeof(T) * num_dimensions);
        }
        ifs.close();
        std::cout << "Finish Reading Data" << endl;
    }

    void load_int(const std::string &file_path)
    {
        std::cout << "Reading Data: " << file_path << std::endl;
        std::ifstream ifs;
        ifs.open(file_path, std::ios::binary);
        assert(ifs.is_open());
        uint32_t N;  // num of points
        ifs.read((char *)&N, sizeof(uint32_t));
        std::cout << "# of points: " << N << std::endl;
        unsigned num_dimensions = 100;
        reset(N, num_dimensions);
        zero();
        ifs.seekg(sizeof(uint32_t), ios::beg);
        std::vector<float> buff(num_dimensions);
        float max= FLT_MIN;
        float min = FLT_MAX;
        //std::cout<<max<<std::endl;
        //std::cout<<min<<std::endl;
        for (unsigned i = 0; i < N; ++i) {
            ifs.read((char *)buff.data(), num_dimensions * sizeof(float));
            std::vector<float> row(num_dimensions);
            for (int d = 0; d < num_dimensions; d++) {
                row[d] = static_cast<float>(buff[d]);
                if(row[d]<min)
                {
                    min=row[d];
                }
                if(row[d]>max)
                {
                    max=row[d];
                }
            }
        }
        //std::cout<<max<<std::endl;
        //std::cout<<min<<std::endl;
        ifs.seekg(sizeof(uint32_t), ios::beg);
        for (unsigned i = 0; i < N; ++i) {
            ifs.read((char *)buff.data(), num_dimensions * sizeof(float));
            std::vector<unsigned short> row(num_dimensions);
            for (int d = 0; d < num_dimensions; d++) {
                row[d] = static_cast<unsigned short>(static_cast<int>((static_cast<float>(buff[d])-min)/(max-min)*8191));
                if(i==100)
                {
                    std::cout<<row[d]<<" ";
                }
            }

            memcpy(&data[stride * i],&row[0],sizeof(T) * num_dimensions);
        }
        ifs.close();
        std::cout << "Finish Reading Data" << endl;
    }
};

template <typename DATA_TYPE, unsigned A = KGRAPH_MATRIX_ALIGN>
class MatrixProxy {
    unsigned rows;
    unsigned cols;      // # elements, not bytes, in a row, 
    size_t stride;    // # bytes in a row, >= cols * sizeof(element)
    uint8_t const *data;
public:
    MatrixProxy (Matrix<DATA_TYPE> const &m)
        : rows(m.size()), cols(m.dim()), stride(m.step()), data(reinterpret_cast<uint8_t const *>(m[0])) {
    }
    unsigned size () const {
        return rows;
    }
    unsigned dim () const {
        return cols;
    }
    DATA_TYPE const *operator [] (unsigned i) const {
        return reinterpret_cast<DATA_TYPE const *>(data + stride * i);
    }
    DATA_TYPE *operator [] (unsigned i) {
        return const_cast<DATA_TYPE *>(reinterpret_cast<DATA_TYPE const *>(data + stride * i));
    }
};

template <typename DATA_TYPE>
class MatrixOracle:public kgraph::IndexOracle{
    MatrixProxy<DATA_TYPE> proxy;
public:
    template <typename MATRIX_TYPE>
    MatrixOracle (MATRIX_TYPE const &m): proxy(m) {
    }
    virtual unsigned size() const{
        return proxy.size();
    }

    virtual dist_type operator()(unsigned i,unsigned j) const{
       //return float_l2sqr_avx512(proxy[i],proxy[j],100);
       //return int_l2sqr_avx512(proxy[i],proxy[j],100);
       return int_EuclideanDistance(proxy[i],proxy[j]);
       //return float_EuclideanDistance(proxy[i],proxy[j]);
    }
    void test()
    {
        std::cout<<std::endl;
        for(int i=0;i<100;i++)
        {
            std::cout<<int(proxy[100][i])<<" ";

        }
        std::cout<<std::endl;
    }
};
